package Stanford::WebApps::SharedEmail::Config;

use strict;
use warnings FATAL => 'all';

use Carp;
use Config::Simple;
require Exporter;
use Readonly;

use base qw(Exporter);

our @EXPORT_OK = qw(%CONFIG);

Readonly my @CONFIG_PATHS => (qw(
    /etc/tools-cgi/common
    /etc/tools-cgi/platform
    /etc/tools-cgi/sharedemail
));
our %CONFIG;

# Read config files
foreach my $config_path (@CONFIG_PATHS) {
    if (-r $config_path) {
        Config::Simple->import_from($config_path, \%CONFIG)
        or croak(  "Error reading '$config_path': " . Config::Simple->error());
    } else {
        croak("Config file not readable at '$config_path': $!");
    }
}

# Make sure we have each of our parameters
Readonly my @CONFIG_PARAMS => (qw(
    tool_ccache
    tool_acctsponsor_server
    tool_ldap_server
    tool_orgperson_server
    tool_platform
    sharedemail_db_host
    sharedemail_db_name
    sharedemail_db_user
    sharedemail_db_passwd_file
    sharedemail_backend_list
    sharedemail_max_forwards
    sharedemail_orgid_db
    sharedemail_wxml_urlbase
    sharedemail_wxml_key
    sharedemail_wxml_cert
));
foreach my $config_param (@CONFIG_PARAMS) {
    if (!defined($CONFIG{$config_param})) {
        croak("Configuration parameter $config_param not set\n");
    }
}

1;

__END__

=pod

=head1 NAME

Stanford::WebApps::SharedEmail::Config - Shared email webapp configuration

=head1 SYNOPSIS

use Stanford::WebApps::SharedEmail::Config qw(%CONFIG);

$ENV{KRB5CCNAME} = $CONFIG{tool_ccache};
# And so on...

=head1 DESCRIPTION

This module reads in common configuration parameters from multiple files, making
them available in the C<%CONFIG> hash, which can be imported directly into
the client's namespace.

Configuration files have a simple form: Parameter name, whitespace, value.

Empty lines are ignored.  Lines starting with C<#> are treated as comments, and are ignored.

=head1 FILES

=over 4

=item /etc/tools-cgi/common

=item /etc/tools-cgi/deptemail

=item /etc/tools-cgi/platform

=back

The list of configuration files is defined in the C<@CONFIG_PATHS> list.
Each configuration file is read into the same namespace.

=head1 OPTIONS

The following configuration options are read from the config files:

=over 4

=item tool_ccache

This is the path to a Kerberos 5 credentials cache, which is used for
Kerberos-authenticated connections to remote hosts.

=item tool_acctsponsor_server

This is the name of the accounts system, used for making remctl calls related to accounts.

=item tool_middleware_server

This is the name of the system used for making remctl calls related to OrgIDs.

=item tool_platform

This is the name of the platform the server is on: Either "prod" or "dev".

=item sharedemail_db_host

This is the name of the MySQL database server.

=item sharedemail_db_name

This is the name of the MySQL database.

=item sharedemail_db_user

This is the username to use for connecting to the MySQL database server.

=item sharedemail_db_passwd-file

This is the path to a file which contains the password for
C<sharedemail_db_user>.  The first line of the file is read and used as
the password.

=item sharedemail_backend_list

The email address, typically a Windows Team address, that should get a
notification when a request is approved.

=item sharedemail_max_forwards

This is the maximum number of forwarding addresses allowed.  The limit is
essentially set by MaIS.

=item sharedemail_orgid_db

This is the name of the Org ID database to read from.  This is either "prod" or "uat".

=back

=head1 AUTHOR

A. Karl Kornel - akkornel@stanford.edu
