package Stanford::WebApps::SharedEmail::Util;

use strict;
use warnings FATAL => 'all';

use Carp;
use Email::Sender::Simple qw(sendmail);
use Email::Simple;
use Email::Simple::Creator;
require Exporter;
use Net::Remctl;
use Readonly;
use Stanford::Directory;
use Stanford::WebApps::SharedEmail::Config qw(%CONFIG);
use Template;
use Try::Tiny;

use base qw(Exporter);

our @EXPORT_OK = qw(lookup_orgid split_entries send_email);

# Create a directory connection for use by lookup_orgid
# Looking up OrgID via LDAP is MUCH faster than via remctl to MaIS
# We'll also let the first query open the connection, in case it isn't needed.
# ALSO, note the anonymous bind!  Otherwise, we won't get org tree access. 8-|
my $DIRECTORY = Stanford::Directory->new();
$DIRECTORY->set(
    ldap_server => $CONFIG{tool_ldap_server},
    mechanism   => undef,
    basedn      => 'cn=organizations, dc=stanford, dc=edu',
);
END { $DIRECTORY->close_directory; }

# lookup_orgid: Given a four-character Org ID, return the name of the
# organization.  If the Org ID doesn't exist, returns false (0).
# Finally, if there are problems, croak.
sub lookup_orgid {
    my ($orgid) = @_;

    # Make sure our OrgID is the right form
    if ($orgid =~ m/\A([a-z]{4})\z/xmsi) {
        $orgid = $1;
    } else {
        return 0;
    }

    # Query LDAP for the OrgID
    my @entries = $DIRECTORY->ldap_query("suAdminID=$orgid", 'displayName');

    # Check for errors
    if ($DIRECTORY->error()) {
        croak("Error looking up OrgID $orgid: " . $DIRECTORY->error());
    }
    if (scalar(@entries) == 0) {
        return 0;
    }
    if (scalar(@entries) > 1) {
        croak("OrgID $orgid found multiple entries");
    }

    # We've got a name, so send it through!
    return $entries[0]->displayName;
}

# split_entries: Given a list of "one X per line", return a list of X
# Takes a string, and returns a list or a list ref.
sub split_entries {
    my ($text) = @_;

    if (!defined($text)) {
        return wantarray ? () : [];
    }
    my @entries = split /^/xms, $text;

    # Remove leading/trailing whitespace and empty lines
    my @output;
    foreach my $entry (@entries) {
        $entry =~ s/\s//xmsg;
        next unless length($entry);
        push @output, lc($entry);
    }

    return wantarray ? @output : \@output;
}

# send_email: This sends an email to someone, using a TT2 template.
# We take the tempate file name, a variables hashref, recipient, and subject
# line.  Will croak if there are any problems.
sub send_email {
    my ($template_file, $template_vars, $recipient, $subject) = @_;

    # $template_vars can be undefined, if the user doesn't have anything
    $template_vars ||= {};

    # Add the platform to the template vars
    $template_vars->{platform} = $CONFIG{tool_platform};

    # Create our template object, and generate the email body
    my $template = Template->new({
        INCLUDE_PATH => [qw(
            /usr/share/webapps/sharedemail/templates
        )],
    });
    my $content;
    $template->process($template_file, $template_vars, \$content)
        or croak ("Error processing $template_file: " . $template->error());

    # Prepare and send the email!
    try {
        my $email = Email::Simple->create(
            header => [
                From    => 'noreply@stanford.edu',
                To      => $recipient,
                Subject => $subject,
            ],
            body => $content,
        );
        sendmail($email);
    } catch {
        my $exception = $_;
        croak(  "Problem sending email '$template_file' to $recipient: "
              . $exception
        );
    };

    return 1;
}

1;
