#!/usr/bin/perl

# sharedemail: 

use strict;
use warnings;

use Carp;
use Config::Simple;
use Readonly;
use Stanford::WebApps::SharedEmail::Config qw(%CONFIG);
use Sys::Syslog qw(:standard :macros);
use Template;
use Try::Tiny;

# Path to common config file
Readonly my $CONFIG_COMMON  = '/etc/tools-cgi/common';

# Path to the webmail config file
Readonly my $CONFIG_WEBMAIL = '/etc/tools-cgi/webmail';

# Load our configuration and set ourselves up

# Start syslog
openlog('webmail', 'ndelay', LOG_LOCAL7);

# Set up our Template object early, in case we need it
my $TEMPLATE = Template->new({
    INCLUDE_PATH => [qw(
        /usr/share/webapps/webmail/templates
    )],
});
our %TEMPLATE_DATA;

# Load our configuration
my %CONFIG;
for my $config_file ($CONFIG_COMMON, $CONFIG_WEBMAIL) {
    Config::Simple->import_from($config_file, \%CONFIG)
    or throw_error(LOG_EMERG,
        "Error reading '$config_file': " . Config::Simple->error(),
        'internal_error'
    );
}
$TEMPLATE_DATA{CONFIG} = \%CONFIG;

# Set our Kerberos credentials cache
local $ENV{KRB5CCNAME} = $CONFIG{tool_ccache};

# Create and set up a directory object
my $LDAP = Stanford::Directory->new();
$LDAP->set(
    ldap_server => $CONFIG{tool_ldap_server},
    mechanism   => 'GSSAPI',
    ticket_file => $CONFIG{tool_ccache},
    basedn      => 'cn=accounts,dc=stanford,dc=edu',
    scope       => 'one',
);

# Look up the email system for a person

# Catch if we didn't get REMOTE_USER, which is very weird.
if (!defined($ENV{REMOTE_USER})) {
    throw_error(LOG_WARNING,
        'Attempted access from someone who doesn\'t have a REMOTE_USER set.',
        'internal_error'
    );
}

# Get the suSeasEmailSystem and suEmailStatus for the account
my @results = $LDAP->ldap_query('uid=' . $ENV{REMOTE_USER},
    'suSeasEmailSystem', 'suEmailStatus',
);

# Handle no entries returned
if (!@results) {
    # An LDAP error is an internal server error
    if ($LDAP->error) {
        throw_error(LOG_EMERG,
            'LDAP error when looking up user: ' . $LDAP->error,
            'internal_error'
        );
    }

    # No entries is a weirder problem
    else {
        throw_error(LOG_ERR,
            'No LDAP entry found for ' . $ENV{REMOTE_USER},
            'missing_account'
        );
    }
}

# Handle multiple entries returned
if (scalar(@results) > 1) {
    throw_error(LOG_ERR,
        'Multiple LDAP entries found for ' . $ENV{REMOTE_USER},
        'internal_error'
    );
}

# Get the user's attributes
my ($email_status) = $results[0]->suEmailStatus;
my ($email_system) = $results[0]->suSeasEmailSystem;

# Catch someone who has a base account
if (   !defined($email_status)
    || ($email_status ne 'active')
) {
    throw_error(LOG_INFO,
        'Failed access by ' . $ENV{REMOTE_USER} . ' who has no email service.',
        'no_email'
    );
}

# Catch someone who is missing suSeasEmailSystem
if (!defined($email_system)) {
    throw_error(LOG_ERR,
        $ENV{REMOTE_USER} . ' is missing suSeasEmailSystem',
        'missing_suSeasEmailSystem'
    );
}

# At this point, we actually have an suSeasEmailSystem!

# Let's see if it is one that we know about.
if (!exists($CONFIG{'services.' . $email_system})) {
    throw_error(LOG_CRIT,
        $ENV{REMOTE_USER} . " has unknown service $email_system",
        'internal_error'
    );
}

# Redirect the user
print CGI->new()->redirect($CONFIG{'services' . $email_system})
or do {
    my $log_message =   'Error redirecting ' . $ENV{REMOTE_USER}
                      . ' to service ' . $email_system . ": $!";
    throw_error(LOG_CRIT, $log_message);
};

# Finally, we're done!
exit 0;


# throw_error: Log an error and exit.  If possible, display a nice page
# for the client.
# Parameters:
#   - log_level: A Sys::Syslog constant.
#   - log_message: A scalar, the message to log to syslog.
#   - template_name: The name of a template to display (no .html.tmpl).
# Does not return.
sub throw_error {
    my ($log_level, $log_message, $template_name) = @_;

    # If we're given a template name, try to process it    
    if (defined($template_name)) {
        my $template_output = q{};
        $TEMPLATE->process($template_name . '.html.tmpl',
                           \%TEMPLATE_DATA,
                           \$template_output
        ) or do {
            $log_message =   $log_message . "\n"
                           . 'ALSO, died with this template error: '
                           . $TEMPLATE->error();
            throw_error($log_level, $log_message);
        };

        # If we got Template output, try to display it
        $template_output = CGI->new()->header() . $template_output;
        print $template_output or do {
            $log_message =   $log_message . "\n"
                           . 'ALSO, was unable to print to STDOUT: '
                           . $!;
            throw_error($log_level, $log_message);
        };

        # If we actually were able to display something, then log
        # to syslog and exit normally.
        syslog($log_level, $log_message);
        exit 0;
    }

    # If we're here, then either there was no template to send, or something
    # really went wrong, so just log to syslog and then die.
    syslog($log_level, $log_message);
    croak($log_message);
}
